@@ -26,7 +26,27 @@ has _writer => (
     predicate => '_has_writer',
 );
 
-sub DEMOLISH { $_[0]->_writer->close if $_[0]->_has_writer }
+has write_fh => (
+  is=>'ro',
+  predicate=>'_has_write_fh',
+  lazy=>1,
+  builder=>'_build_write_fh',
+);
+
+sub _build_write_fh {
+  my $self = shift;
+  $self->_context->finalize_headers unless
+    $self->finalized_headers;
+  $self->_writer;
+};
+
+sub DEMOLISH {
+  my $self = shift;
+  return if $self->_has_write_fh;
+  if($self->_has_writer) {
+    $self->_writer->close
+  }
+}
 
 has cookies   => (is => 'rw', default => sub { {} });
 has body      => (is => 'rw', default => undef);
@@ -92,6 +112,38 @@ sub finalize_headers {
     return;
 }
 
+sub from_psgi_response {
+    my ($self, $psgi_res) = @_;
+    if(ref $psgi_res eq 'ARRAY') {
+        my ($status, $headers, $body) = @$psgi_res;
+        $self->status($status);
+        $self->headers(HTTP::Headers->new(@$headers));
+        if(ref $body eq 'ARRAY') {
+          $self->body(join '', grep defined, @$body);
+        } else {
+          $self->body($body);
+        }
+    } elsif(ref $psgi_res eq 'CODE') {
+        $psgi_res->(sub {
+            my $response = shift;
+            my ($status, $headers, $maybe_body) = @$response;
+            $self->status($status);
+            $self->headers(HTTP::Headers->new(@$headers));
+            if($maybe_body) {
+                if(ref $maybe_body eq 'ARRAY') {
+                  $self->body(join '', grep defined, @$maybe_body);
+                } else {
+                  $self->body($maybe_body);
+                }
+            } else {
+                return $self->write_fh;
+            }
+        });  
+     } else {
+        die "You can't set a Catalyst response from that, expect a valid PSGI response";
+    }
+}
+
 =head1 NAME
 
 Catalyst::Response - stores output responding to the current client request
@@ -246,6 +298,40 @@ $res->code is an alias for this, to match HTTP::Response->code.
 
 Writes $data to the output stream.
 
+=head2 $res->write_fh
+
+Returns a PSGI $writer object that has two methods, write and close.  You can
+close over this object for asynchronous and nonblocking applications.  For
+example (assuming you are using a supporting server, like L<Twiggy>
+
+    package AsyncExample::Controller::Root;
+
+    use Moose;
+
+    BEGIN { extends 'Catalyst::Controller' }
+
+    sub prepare_cb {
+      my $write_fh = pop;
+      return sub {
+        my $message = shift;
+        $write_fh->write("Finishing: $message\n");
+        $write_fh->close;
+      };
+    }
+
+    sub anyevent :Local :Args(0) {
+      my ($self, $c) = @_;
+      my $cb = $self->prepare_cb($c->res->write_fh);
+
+      my $watcher;
+      $watcher = AnyEvent->timer(
+        after => 5,
+        cb => sub {
+          $cb->(scalar localtime);
+          undef $watcher; # cancel circular-ref
+        });
+    }
+
 =head2 $res->print( @data )
 
 Prints @data to the output stream, separated by $,.  This lets you pass
@@ -255,6 +341,33 @@ the response object to functions that want to write to an L<IO::Handle>.
 
 Writes headers to response if not already written
 
+=head2 from_psgi_response
+
+Given a PSGI response (either three element ARRAY reference OR coderef expecting
+a $responder) set the response from it.
+
+Properly supports streaming and delayed response and / or async IO if running
+under an expected event loop.
+
+Example:
+
+    package MyApp::Web::Controller::Test;
+
+    use base 'Catalyst::Controller';
+    use Plack::App::Directory;
+
+
+    my $app = Plack::App::Directory->new({ root => "/path/to/htdocs" })
+      ->to_app;
+
+    sub myaction :Local Args {
+      my ($self, $c) = @_;
+      $c->res->from_psgi_response($app->($self->env));
+    }
+
+Please note this does not attempt to map or nest your PSGI application under
+the Controller and Action namespace or path.  
+
 =head2 DEMOLISH
 
 Ensures that the response is flushed and closed at the end of the
